/*	$NetBSD$	*/

/*-
 * Copyright 2009 Brett Lymn <blymn@NetBSD.org>
 *
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <curses.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include "returns.h"
#include "slave.h"
#include "command_table.h"
#include "test_commands.h"

extern int cmdpipe[2];
extern int slvpipe[2];

static void report_type(returns_enum_t);

/*
 * Match the passed command string and execute the associated test
 * function.
 */
void
command_execute(char *func, int nargs, char **args)
{
	size_t i;

	i = 0;
	while (i < ncmds) {
		if (strcasecmp(func, commands[i].name) == 0) {
			/* matched function */
			commands[i].func(nargs, args);
			return;
		}
		i++;
	}

	report_status("UNKNOWN_FUNCTION");
}

/*
 * Report an pointer value back to the director
 */
void
report_ptr(void *ptr)
{
	char *string;

	asprintf(&string, "%td", ptr);
	report_status(string);
	free(string);
}

/*
 * Report an integer value back to the director
 */
void
report_int(int value)
{
	char *string;

	asprintf(&string, "%d", value);
	report_status(string);
	free(string);
}

/*
 * Report either an ERR or OK back to the director
 */
void
report_return(int status)
{
	if (status == ERR)
		report_type(ret_err);
	else if (status == OK)
		report_type(ret_ok);
	else
		report_status("INVALID_RETURN");
}

/*
 * Report the type back to the director via the command pipe
 */
static void
report_type(returns_enum_t return_type)
{
	int type;

	type = return_type;
	if (write(slvpipe[WRITE_PIPE], &type, sizeof(int)) < 0) {
		perror("command pipe write for status type failed");
		exit(1);
	}

}

/*
 * Report the status back to the director via the command pipe
 */
void
report_status(char *status)
{
	int len, type;

	len = strlen(status);

	type = ret_string;
	if (write(slvpipe[WRITE_PIPE], &type, sizeof(int)) < 0) {
		perror("command pipe write for status type failed");
		exit(1);
	}

	if (write(slvpipe[WRITE_PIPE], &len, sizeof(int)) < 0) {
		perror("command pipe write for status length failed");
		exit(1);
	}

	if (write(slvpipe[WRITE_PIPE], status, len) < 0) {
		perror("command pipe write of status data failed");
		exit(1);
	}
}

/*
 * Report a string of chtype back to the director via the command pipe.
 */
void
report_nstr(chtype *string)
{
	int len, type;
	chtype *p;

	len = 0;
	p = string;

	while ((*p++ & __CHARTEXT) != 0) {
		len++;
	}

	type = ret_byte;
	if (write(slvpipe[WRITE_PIPE], &type, sizeof(int)) < 0) {
		perror("report_nstr: command pipe write for status type failed");
		exit(1);
	}

	if (write(slvpipe[WRITE_PIPE], &len, sizeof(int)) < 0) {
		perror("report_nstr: command pipe write for status length failed");
		exit(1);
	}

	if (write(slvpipe[WRITE_PIPE], string, len) < 0) {
		perror("report_nstr: command pipe write of status data failed");
		exit(1);
	}
}

/*
 * Check the number of args we received are what we expect.  Return an
 * error if they do not match.
 */
int
check_arg_count(int nargs, int expected)
{
	if (nargs != expected) {
		report_status("INCORRECT_ARGUMENT_NUMBER");
		return(1);
	}

	return(0);
}
